import os
import subprocess
import uuid

import mlflow

import docker
from gaiaflow.constants import (DEFAULT_MINIO_AWS_ACCESS_KEY_ID,
                                DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY,
                                DEFAULT_MLFLOW_TRACKING_URI, BaseAction)
from gaiaflow.managers.base_manager import BaseGaiaflowManager


class MlflowModelManager(BaseGaiaflowManager):
    def __init__(
        self,
        registry_uri=None,
        action: BaseAction = None,
        force_new: bool = False,
        prune: bool = False,
        local: bool = False,
        **kwargs,
    ):
        self.docker_client = docker.from_env()
        self.registry_uri = registry_uri.rstrip("/") if registry_uri else None
        self.local = local

        # TODO: not optimal. Use ContextManager to restore the env vars
        #  Move them to the actual methods where they are needed.
        os.environ["MLFLOW_TRACKING_URI"] = (
            os.getenv("MLFLOW_TRACKING_URI") or "https://localhost:5000"
        )
        os.environ["MLFLOW_S3_ENDPOINT_URL"] = (
            os.getenv("MLFLOW_S3_ENDPOINT_URL") or "https://localhost:9000"
        )
        os.environ["AWS_ACCESS_KEY_ID"] = (
            os.getenv("AWS_ACCESS_KEY_ID") or DEFAULT_MINIO_AWS_ACCESS_KEY_ID
        )
        os.environ["AWS_SECRET_ACCESS_KEY"] = (
            os.getenv("AWS_SECRET_ACCESS_KEY") or DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY
        )

        mlflow_tracking_uri = (
            os.getenv("MLFLOW_TRACKING_URI") or DEFAULT_MLFLOW_TRACKING_URI
        )
        mlflow.set_tracking_uri(mlflow_tracking_uri)

        # This Manager does not need the paths. But it expects that either
        # the model exists in S3 if s3_uri is used or a live instance of MLFlow
        # running if run_id is used
        super().__init__(
            gaiaflow_path="",
            user_project_path="",
            action=action,
            force_new=force_new,
            prune=prune,
            **kwargs,
        )

    def _get_model_uri(self, model_uri=None, run_id=None):
        if model_uri:
            return model_uri
        elif run_id:
            return f"runs:/{run_id}/model"
        else:
            raise ValueError("Either model_uri or run_id must be provided.")

    # TODO: Use this instead
    # @classmethod
    # def run(cls, action, **kwargs):
    #     params = kwargs.get("params", {})
    #     model_uri = params.get("model_uri")
    #     run_id = params.get("run_id")
    #     image_name = params.get("image_name")
    #     enable_mlserver = params.get("enable_mlserver", True)
    #     manag = MlflowModelManager(model_uri, run_id, image_name,
    #                               enable_mlserver)
    #     environment = create_env()
    #     with set_env_cm(environment):
    #         if action == BaseAction.START:
    #             manag.start()

    def start(self, **kwargs):
        print("kwargs inside start, **kwargs", kwargs)
        params = kwargs.get("params", {})

        model_uri = params.get("model_uri")
        run_id = params.get("run_id")
        image_name = params.get("image_name")
        enable_mlserver = params.get("enable_mlserver", True)

        self.build_model_docker_image(model_uri, run_id, image_name, enable_mlserver)
        self.run_container(image_name)

    def stop(self, **kwargs):
        print("kwargs inside stop, **kwargs", kwargs)
        params = kwargs.get("params", {})

        container_id_or_name = params.get("container_id_or_name")
        self.stop_and_remove_container(container_id_or_name)

    def cleanup(self, **kwargs):
        print("kwargs inside cleanup, **kwargs", kwargs)
        params = kwargs.get("params", {})

        container_id_or_name = params.get("container_id_or_name")
        purge = params.get("purge")
        self.stop_and_remove_container(container_id_or_name, purge)

    def build_model_docker_image(
        self, model_uri=None, run_id=None, image_name=None, enable_mlserver=True
    ):
        model_uri_final = self._get_model_uri(model_uri, run_id)

        unique_tag = str(uuid.uuid4())[:8]
        image_name = image_name or f"mlflow-model:{unique_tag}"

        full_image_name = (
            f"{self.registry_uri}/{image_name}" if self.registry_uri else image_name
        )

        print(f"Building Docker image from {model_uri_final} as {full_image_name}...")

        mlflow.models.build_docker(
            model_uri=model_uri_final,
            name=full_image_name,
            enable_mlserver=enable_mlserver,
        )

        return full_image_name

    def run_container(self, image_name, port=8246):
        container = self.docker_client.containers.run(
            image_name,
            detach=True,
            ports={"8080/tcp": port},
            name=f"mlflow_model_{uuid.uuid4().hex[:6]}",
        )
        print(f"Container started: {container.name} on port {port}")
        return container

    def stop_and_remove_container(self, container_id_or_name, purge=False):
        try:
            container = self.docker_client.containers.get(container_id_or_name)
            print(f"Stopping container: {container.name}")
            container.stop()
            container.remove()
            print("Container stopped and removed.")
            if purge:
                image_name = container.image.tags[0]
                self.docker_client.images.remove(image=image_name, force=True)
                print(f"Image {image_name} removed.")
        except docker.context.api.errors.NotFound:
            print("Container not found.")

    def push_image_to_registry(self, image_name):
        if not self.registry_uri:
            raise ValueError("No registry_uri provided.")
        print(f"Pushing image {image_name} to registry {self.registry_uri}...")
        subprocess.run(["docker", "push", image_name], check=True)
        print("Image pushed.")

    def generate_k8s_yaml(self):
        print(" Generating Kubernetes deployment YAML...")
        raise NotImplementedError()

    def deploy_to_kubernetes(self, k8s_yaml_path):
        print(f"Deploying to Kubernetes from {k8s_yaml_path}...")
        subprocess.run(["kubectl", "apply", "-f", k8s_yaml_path], check=True)
        print("Deployment applied.")


#
# @contextmanager
# def set_env_cm(**new_env: str | None) -> Generator[dict[str, str | None]]:
#     """Run the code in the block with a new environment `new_env`."""
#     restore_env = set_env(**new_env)
#     try:
#         yield new_env
#     finally:
#         restore_env()
#
#
# def set_env(**new_env: str | None) -> Callable[[], None]:
#     """
#     Set the new environment in `new_env` and return a no-arg
#     function to restore the old environment.
#     """
#     old_env = {k: os.environ.get(k) for k in new_env.keys()}
#
#     def restore_env():
#         for ko, vo in old_env.items():
#             if vo is not None:
#                 os.environ[ko] = vo
#             elif ko in os.environ:
#                 del os.environ[ko]
#
#     for kn, vn in new_env.items():
#         if vn is not None:
#             os.environ[kn] = vn
#         elif kn in os.environ:
#             del os.environ[kn]
#
#     return restore_env



from importlib.metadata import version

import typer

from gaiaflow.cli.commands import minikube, mlops

pkg_version = version("gaiaflow")

app = typer.Typer(
    add_completion=False,
    help=(
        "Gaiaflow CLI is a manager tool "
        "to allow you to create/destroy your local MLOps infrastructure that "
        "mirrors closely to the actual production systems for easier and faster "
        "local development, iteration and testing."
    ),
)

app.add_typer(mlops.app, name="dev", help="Manage Gaiaflow development services.")
app.add_typer(
    minikube.app,
    name="prod-local",
    help=(
        "Manage Gaiaflow production-like services. Start this once "
        "you have developed your package and tested your workflow in the dev mode."
    ),
)
# app.add_typer(
#     mlflow_model.app,
#     name="model",
#     help=(
#         "Manage MLFlow models by either deploying or un-deploying them. "
#         "NOTE:Currently only local deployments supported."
#     ),
# )

if __name__ == "__main__":
    app()



from enum import Enum

from .operators import (DevTaskOperator, DockerTaskOperator,
                        ProdLocalTaskOperator, ProdTaskOperator)


class GaiaflowMode(Enum):
    DEV = "dev"
    DEV_DOCKER = "dev_docker"
    PROD_LOCAL = "prod_local"
    PROD = "prod"


OPERATOR_MAP = {
    GaiaflowMode.DEV: DevTaskOperator,
    GaiaflowMode.DEV_DOCKER: DockerTaskOperator,
    GaiaflowMode.PROD_LOCAL: ProdLocalTaskOperator,
    GaiaflowMode.PROD: ProdTaskOperator,
}


def create_task(
    task_id: str,
    func_path: str,
    func_kwargs: dict | None = None,
    func_args: list | None = None,
    image: str | None = None,
    mode: str = "dev",
    secrets: list | None = None,
    env_vars: dict | None = None,
    retries: int = 3,
    dag=None,
):
    """It is a high-level abstraction on top of Apache Airflow operators.

    It allows you to define tasks for your DAGs in a uniform, environment-aware
    way, without worrying about which Airflow operator to use for each execution mode.
    """
    try:
        gaiaflow_mode: GaiaflowMode = GaiaflowMode(mode)
    except ValueError:
        raise ValueError(
            f"env must be one of {[e.value for e in GaiaflowMode]}, got '{mode}'"
        )

    func_args = func_args or []
    func_kwargs = func_kwargs or {}
    if env_vars is None:
        env_vars = {}

    dag_params = getattr(dag, "params", {}) if dag else {}

    operator_cls = OPERATOR_MAP.get(gaiaflow_mode)
    if not operator_cls:
        raise ValueError(f"No task creation operator defined for {gaiaflow_mode}")

    operator = operator_cls(
        task_id=task_id,
        func_path=func_path,
        func_args=func_args,
        func_kwargs=func_kwargs,
        image=image,
        secrets=secrets,
        env_vars=env_vars,
        retries=retries,
        params=dag_params,
        mode=gaiaflow_mode,
    )

    return operator.create_task()



from dataclasses import dataclass
from enum import Enum
from pathlib import Path


@dataclass(frozen=True)
class Action:
    name: str


class BaseAction:
    START = Action("start")
    STOP = Action("stop")
    RESTART = Action("restart")
    CLEANUP = Action("cleanup")


class ExtendedAction:
    DOCKERIZE = Action("dockerize")
    CREATE_CONFIG = Action("create_config")
    CREATE_SECRET = Action("create_secret")


GAIAFLOW_CONFIG_DIR = Path.home() / ".gaiaflow"
GAIAFLOW_CONFIG_DIR.mkdir(exist_ok=True)
GAIAFLOW_STATE_FILE = GAIAFLOW_CONFIG_DIR / "state.json"


class Service(str, Enum):
    airflow = "airflow"
    mlflow = "mlflow"
    minio = "minio"
    jupyter = "jupyter"
    all = "all"


AIRFLOW_SERVICES = [
    "airflow-apiserver",
    "airflow-scheduler",
    "airflow-init",
    "airflow-dag-processor",
    "airflow-triggerer",
    "postgres-airflow",
]

MLFLOW_SERVICES = ["mlflow", "postgres-mlflow"]

MINIO_SERVICES = ["minio", "minio_client"]

DEFAULT_MINIO_AWS_ACCESS_KEY_ID = "minio"
DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY = "minio123"
DEFAULT_MLFLOW_TRACKING_URI = "http://localhost:5000"

# TODO: Talk with Tejas/Norman (currently contains random values)
RESOURCE_PROFILES = {
    "low": {
        "request_cpu": "250m",
        "limit_cpu": "500m",
        "request_memory": "512Mi",
        "limit_memory": "1Gi",
        "limit_gpu": "0",
    },
    "medium": {
        "request_cpu": "500m",
        "limit_cpu": "1",
        "request_memory": "1Gi",
        "limit_memory": "2Gi",
        "limit_gpu": "0",
    },
    "high": {
        "request_cpu": "1",
        "limit_cpu": "2",
        "request_memory": "2Gi",
        "limit_memory": "4Gi",
        "limit_gpu": "0.5",
    },
    "ultra": {
        "request_cpu": "2",
        "limit_cpu": "4",
        "request_memory": "4Gi",
        "limit_memory": "8Gi",
        "limit_gpu": "1",
    },
}

DEFAULT_IMAGE_NAME = "user-image:v1"



import json
import subprocess
import sys
import tempfile
from datetime import datetime
from pathlib import Path

import fsspec
import typer

from gaiaflow.constants import GAIAFLOW_STATE_FILE

fs = fsspec.filesystem("file")


def get_gaialfow_version() -> str:
    try:
        from importlib.metadata import version

        return version("gaiaflow")
    except Exception:
        print("Package not installed. Getting version from the pyproject.toml")
        import tomllib

        pyproject = tomllib.loads(Path("pyproject.toml").read_text())
        return pyproject["project"]["version"]


def find_python_packages(base_path: Path):
    outer_packages = []

    for child in base_path.iterdir():
        if child.is_dir() and (child / "__init__.py").exists():
            outer_packages.append(child.name)

    return outer_packages


def log_info(message: str):
    print(f"\033[0;34m[{datetime.now().strftime('%H:%M:%S')}]\033[0m {message}")


def log_error(message: str):
    print(f"\033[0;31mERROR:\033[0m {message}", file=sys.stderr)


def run(command: list, error_message: str, env=None):
    try:
        subprocess.call(command, env=env)
    except Exception:
        log_error(error_message)
        raise


def handle_error(message: str):
    log_error(f"Error: {message}")
    sys.exit(1)


def get_state_file() -> Path:
    return GAIAFLOW_STATE_FILE


def save_project_state(project_path: Path, gaiaflow_path: Path):
    state_file = get_state_file()
    try:
        if state_file.exists():
            with state_file.open("r") as f:
                state = json.load(f)
        else:
            state = {}
    except (json.JSONDecodeError, FileNotFoundError):
        state = {}

    project_path_str = str(project_path)
    gaiaflow_path_str = str(gaiaflow_path)
    keys_to_delete = [
        k
        for k, v in state.items()
        if v.get("project_path") == project_path_str and k != gaiaflow_path_str
    ]
    for k in keys_to_delete:
        del state[k]

    state[gaiaflow_path_str] = {
        "project_path": project_path_str,
    }

    with open(state_file, "w") as f:
        json.dump(state, f, indent=2)


def load_project_state() -> dict | None:
    state_file = get_state_file()
    print("state_file", state_file)
    if not state_file.exists():
        return None

    try:
        with open(state_file, "r") as f:
            return json.load(f)
    except (json.JSONDecodeError, FileNotFoundError):
        return None


def check_structure(base_path: Path, structure: dict) -> bool:
    for name, sub in structure.items():
        if name == "_files_":
            for file_name in sub:
                file_path = base_path / file_name
                if not file_path.exists():
                    typer.echo(f"Missing file: {file_path}", err=True)
                    return False
            continue

        path = base_path / name
        if not path.exists() or not path.is_dir():
            typer.echo(f"Missing folder: {path}", err=True)
            return False

        if isinstance(sub, dict):
            if not check_structure(path, sub):
                return False
        elif isinstance(sub, list):
            for item in sub:
                item_path = path / item
                if not item_path.exists():
                    typer.echo(f"Missing file: {item_path}", err=True)
                    return False
        else:
            raise TypeError("Structure values must be dict or list")
    return True


def gaiaflow_path_exists_in_state(gaiaflow_path: Path, check_fs: bool = True) -> bool:
    REQUIRED_STRUCTURE = {
        "docker_stuff": {
            "docker-compose": [
                "docker-compose.yml",
                "docker-compose-minikube-network.yml",
                "entrypoint.sh",
            ],
            "airflow": ["Dockerfile"],
            "mlflow": ["Dockerfile", "requirements.txt"],
            "user-package": ["Dockerfile"],
            "_files_": ["kube_config_inline"],
        }
    }
    state = load_project_state()
    if not state:
        return False

    key = str(gaiaflow_path)
    if key not in state:
        return False

    if check_fs:
        if not gaiaflow_path.exists():
            typer.echo(
                f"Gaiaflow path exists in state but not on disk: {gaiaflow_path}",
                err=True,
            )
            return False

        if not check_structure(gaiaflow_path, REQUIRED_STRUCTURE):
            return False

    return True


def delete_project_state(gaiaflow_path: Path):
    state_file = get_state_file()
    print("state_file", state_file)
    if not state_file.exists():
        log_error(
            "State file not found at ~/.gaiaflow/state.json. Please run the services."
        )
        return

    try:
        with open(state_file, "r") as f:
            state = json.load(f)

        print("found!", state.get("gaiaflow_path"), state)
        key = str(gaiaflow_path)
        if key in state:
            del state[key]
            with open(state_file, "w") as f:
                json.dump(state, f, indent=2)
    except (json.JSONDecodeError, FileNotFoundError, Exception):
        raise


def parse_key_value_pairs(pairs: list[str]) -> dict:
    data = {}
    for pair in pairs:
        if "=" not in pair:
            raise typer.BadParameter(f"Invalid format: '{pair}'. Expected key=value.")
        key, value = pair.split("=", 1)
        data[key] = value
    return data


def create_directory(dir_name):
    if not fs.exists(dir_name):
        try:
            fs.makedirs(dir_name, exist_ok=True)
            log_info(f"Created directory: {dir_name}")
        except Exception as e:
            handle_error(f"Failed to create {dir_name} directory: {e}")
    else:
        log_info(f"Directory {dir_name} already exists")

    set_permissions(dir_name)


def set_permissions(path, mode=0o777):
    try:
        fs.chmod(path, mode)
        log_info(f"Set permissions for {path}")
    except Exception:
        log_info(f"Warning: Could not set permissions for {path}")


def create_gaiaflow_context_path(project_path: Path) -> tuple[Path, Path]:
    user_project_path = Path(project_path).resolve()
    if not user_project_path.exists():
        raise FileNotFoundError(f"{user_project_path} not found")
    version = get_gaialfow_version()
    # project_name = str(user_project_path).split("/")[-1]
    project_name = user_project_path.name
    tmp_dir = Path(tempfile.gettempdir())
    gaiaflow_path =  tmp_dir / f"gaiaflow-{version}-{project_name}"

    return gaiaflow_path, user_project_path


def convert_crlf_to_lf(file_path: str):
    """
    Converts a file from Windows-style CRLF line endings to Unix-style LF line endings.

    Args:
        file_path (str): Path to the file to convert.
    """
    with open(file_path, "rb") as f:
        content = f.read()

    # Replace CRLF (\r\n) with LF (\n)
    new_content = content.replace(b"\r\n", b"\n")

    with open(file_path, "wb") as f:
        f.write(new_content)

    print(f"Converted {file_path} to LF line endings.")


def is_wsl() -> bool:
    try:
        with open("/proc/version", "r") as f:
            return "microsoft" in f.read().lower()
    except FileNotFoundError:
        return False


import typer

from gaiaflow.constants import BaseAction
from gaiaflow.managers.mlflow_model_manager import MlflowModelManager

app = typer.Typer()


@app.command(
    help="Deploy a Mlflow model. NOTE: Currently only local model "
    "deployment is provided."
)
def start(
    model_uri: str = typer.Option(
        "",
        help="S3 URI of the model to serve. "
        "This or run_id has to be "
        "provided. If you provide both, "
        "model_uri takes precedence.",
    ),
    run_id: str = typer.Option(
        "",
        help="MLflow run ID of your model. This "
        "or model_uri has to be provided. If you provide both, "
        "model_uri takes precedence.",
    ),
    image_name: str = typer.Option(..., help="Docker image name"),
    enable_mlserver: bool = typer.Option(True, help="Enable MLServer"),
):
    if not (run_id or model_uri):
        typer.echo("Please provide either run_id or model_uri.")
    kwargs = {
        "params": {
            "model_uri": model_uri,
            "run_id": run_id,
            "image_name": image_name,
            "enable_mlserver": enable_mlserver,
        }
    }
    typer.echo(f"kwargs::{kwargs}")
    MlflowModelManager(action=BaseAction.START, **kwargs)


@app.command(
    help="Stop a deployed Mlflow model container. NOTE: Currently "
    "only local model deployment is provided."
)
def stop(
    container_id_or_name: str = typer.Option(..., help="Container ID or name"),
):
    kwargs = {
        "params": {
            "container_id_or_name": container_id_or_name,
        }
    }
    typer.echo(f"kwargs::{kwargs}")
    MlflowModelManager(action=BaseAction.STOP, **kwargs)


@app.command(
    help="Cleanup a deployed Mlflow model container. NOTE: Currently "
    "only local model deployment is provided."
)
def cleanup(
    container_id_or_name: str = typer.Option(..., help="Container ID or name"),
    purge: str = typer.Option(..., "-p", help="Remove the Deployed MLFlow model image"),
):
    kwargs = {
        "params": {
            "container_id_or_name": container_id_or_name,
            "purge": purge,
        }
    }
    typer.echo(f"kwargs::{kwargs}")
    MlflowModelManager(action=BaseAction.STOP, **kwargs)



import re
import subprocess

from kubernetes.client import V1EnvFromSource, V1SecretReference


def inject_params_as_env_vars(params: dict) -> dict:
    return {f"PARAMS_{k.upper()}": f"{{{{ params.{k} }}}}" for k in params}


def build_env_from_secrets(secrets: list[str]) -> list[V1EnvFromSource]:
    return [
        V1EnvFromSource(secret_ref=V1SecretReference(name=secret)) for secret in secrets
    ]


def docker_network_gateway() -> str | None:
    network_name = "airflow"
    try:
        result = subprocess.run(
            ["docker", "network", "inspect", network_name],
            check=True,
            capture_output=True,
            text=True,
        )

        for line in result.stdout.splitlines():
            if "Gateway" in line:
                match = re.search(r'"Gateway"\s*:\s*"([^"]+)"', line)
                if match:
                    print(f"Docker network Gateway for Minikube is - "
                          f"{match.group(1)}")
                    return match.group(1)
        print("Is your minikube cluster running? Please run and try again.")
        return None

    except subprocess.CalledProcessError as e:
        print(f"Error running docker network inspect: {e}")
        return None
    except FileNotFoundError:
        print("Docker command not found. Is Docker installed and in your PATH?")
        return None



from pathlib import Path
from types import SimpleNamespace
from typing import List

import fsspec
import typer

from gaiaflow.constants import Service, DEFAULT_IMAGE_NAME

app = typer.Typer()
fs = fsspec.filesystem("file")


def load_imports():
    from gaiaflow.constants import BaseAction, ExtendedAction
    from gaiaflow.managers.mlops_manager import MlopsManager
    from gaiaflow.managers.minikube_manager import MinikubeManager
    from gaiaflow.managers.utils import (create_gaiaflow_context_path,
                                         gaiaflow_path_exists_in_state,
                                         save_project_state)

    return SimpleNamespace(
        BaseAction=BaseAction,
        ExtendedAction=ExtendedAction,
        MlopsManager=MlopsManager,
        MinikubeManager=MinikubeManager,
        create_gaiaflow_context_path=create_gaiaflow_context_path,
        gaiaflow_path_exists_in_state=gaiaflow_path_exists_in_state,
        save_project_state=save_project_state,
    )


@app.command(help="Start Gaiaflow development services")
def start(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    force_new: bool = typer.Option(
        False,
        "--force-new",
        "-f",
        help="If you need a fresh gaiaflow installation. "
        "NOTE. It only removes the current version of Gaiaflow. If you need "
        "to remove the docker related stuff, use the cleanup --prune "
        "command",
    ),
    service: List[Service] = typer.Option(
        ["all"],
        "--service",
        "-s",
        help="Services to manage. Use multiple --service flags, or leave empty to run all.",
    ),
    cache: bool = typer.Option(False, "--cache", "-c", help="Use Docker cache"),
    jupyter_port: int = typer.Option(
        8895, "--jupyter-port", "-j", help="Port for JupyterLab"
    ),
    delete_volume: bool = typer.Option(
        False, "--delete-volume", "-v", help="Delete volumes on shutdown"
    ),
    docker_build: bool = typer.Option(
        False, "--docker-build", "-b", help="Force Docker image build"
    ),
):
    imports = load_imports()
    typer.echo(f"Selected Gaiaflow services: {service}")
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        imports.save_project_state(user_project_path, gaiaflow_path)
    else:
        typer.echo(
            f"Gaiaflow project already exists at {gaiaflow_path}. Skipping "
            f"saving to the state"
        )

    if service != ["all"]:
        for s in service:
            typer.echo(f"Running start on {s}...")
            imports.MlopsManager.run(
                gaiaflow_path=gaiaflow_path,
                user_project_path=user_project_path,
                force_new=force_new,
                action=imports.BaseAction.START,
                service=s,
                cache=cache,
                jupyter_port=jupyter_port,
                delete_volume=delete_volume,
                docker_build=docker_build,
            )
    else:
        typer.echo("Running start with all services")
        imports.MlopsManager.run(
            gaiaflow_path=gaiaflow_path,
            user_project_path=user_project_path,
            force_new=force_new,
            action=imports.BaseAction.START,
            service=Service.all,
            cache=cache,
            jupyter_port=jupyter_port,
            delete_volume=delete_volume,
            docker_build=docker_build,
        )


@app.command(help="Stop Gaiaflow development services")
def stop(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    service: List[Service] = typer.Option(
        ["all"],
        "--service",
        "-s",
        help="Services to manage. Use multiple --service flags, or leave empty to run all.",
    ),
    delete_volume: bool = typer.Option(
        False, "--delete-volume", "-v", help="Delete volumes on shutdown"
    ),
):
    """"""
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
    if service != ["all"]:
        for s in service:
            typer.echo(f"Stopping service:  {s}")
            imports.MlopsManager.run(
                gaiaflow_path=Path(gaiaflow_path),
                user_project_path=Path(user_project_path),
                action=imports.BaseAction.STOP,
                service=s,
                delete_volume=delete_volume,
            )
    else:
        typer.echo("Stopping all services")
        imports.MlopsManager.run(
            gaiaflow_path=Path(gaiaflow_path),
            user_project_path=Path(user_project_path),
            service=Service.all,
            action=imports.BaseAction.STOP,
            delete_volume=delete_volume,
        )


@app.command(help="Restart Gaiaflow development services")
def restart(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    force_new: bool = typer.Option(
        False,
        "--force-new",
        "-f",
        help="If you need a "
        "fresh gaiaflow installation. "
        "NOTE. Currently it only removes "
        "the current version of Gaiaflow.",
    ),
    service: List[Service] = typer.Option(
        ["all"],
        "--service",
        "-s",
        help="Services to manage. Use multiple --service flags, or leave empty to run all.",
    ),
    delete_volume: bool = typer.Option(
        False, "--delete-volume", "-v", help="Delete volumes on shutdown"
    ),
    cache: bool = typer.Option(False, "--cache", "-c", help="Use Docker cache"),
    jupyter_port: int = typer.Option(
        8895, "--jupyter-port", "-j", help="Port for JupyterLab"
    ),
    docker_build: bool = typer.Option(
        False, "--docker-build", "-b", help="Force Docker image build"
    ),
):
    """"""
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
    if service != ["all"]:
        for s in service:
            typer.echo(f"Stopping service:  {s}")
            imports.MlopsManager.run(
                gaiaflow_path=Path(gaiaflow_path),
                user_project_path=Path(user_project_path),
                force_new=force_new,
                action=imports.BaseAction.RESTART,
                service=s,
                cache=cache,
                jupyter_port=jupyter_port,
                delete_volume=delete_volume,
                docker_build=docker_build,
            )
    else:
        typer.echo("Stopping all services")
        imports.MlopsManager.run(
            gaiaflow_path=Path(gaiaflow_path),
            user_project_path=Path(user_project_path),
            force_new=force_new,
            action=imports.BaseAction.RESTART,
            cache=cache,
            jupyter_port=jupyter_port,
            delete_volume=delete_volume,
            docker_build=docker_build,
            service=Service.all,
        )


@app.command(
    help="Clean Gaiaflow development services. This will remove the "
    "gaiaflow static context directory from the /tmp folder and "
    "also remove the state for this project."
)
def cleanup(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    prune: bool = typer.Option(
        False, "--prune", help="Prune Docker image, network and cache"
    ),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
    imports.MlopsManager.run(
        gaiaflow_path=Path(gaiaflow_path),
        user_project_path=Path(user_project_path),
        action=imports.BaseAction.CLEANUP,
        prune=prune,
    )



@app.command(help="Containerize your package into a docker image locally.")
def dockerize(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    image_name: str = typer.Option(DEFAULT_IMAGE_NAME, "--image-name", "-i",
                                   help=("Name of your image.")),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        imports.save_project_state(user_project_path, gaiaflow_path)
    else:
        typer.echo(
            f"Gaiaflow project already exists at {gaiaflow_path}. Skipping "
            f"saving to the state"
        )

    typer.echo("Running dockerize")
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.ExtendedAction.DOCKERIZE,
        local=True,
        image_name=image_name
    )



# TODO: To let the user update the current infra with new local packages or
#  mounts as they want it.
# def update():



from pathlib import Path
from types import SimpleNamespace

import fsspec
import typer

from gaiaflow.constants import DEFAULT_IMAGE_NAME

app = typer.Typer()
fs = fsspec.filesystem("file")


def load_imports():
    from gaiaflow.constants import BaseAction
    from gaiaflow.managers.minikube_manager import (ExtendedAction,
                                                    MinikubeManager)
    from gaiaflow.managers.utils import (create_gaiaflow_context_path,
                                         gaiaflow_path_exists_in_state,
                                         parse_key_value_pairs)

    return SimpleNamespace(
        BaseAction=BaseAction,
        ExtendedAction=ExtendedAction,
        MinikubeManager=MinikubeManager,
        create_gaiaflow_context_path=create_gaiaflow_context_path,
        gaiaflow_path_exists_in_state=gaiaflow_path_exists_in_state,
        parse_key_value_pairs=parse_key_value_pairs,
    )


@app.command(help="Start Gaiaflow production-like services.")
def start(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    force_new: bool = typer.Option(
        False,
        "--force-new",
        "-f",
        help="Set this to true if you need a "
        "fresh production-like environment installation. ",
    ),
):
    """"""
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.BaseAction.START,
        force_new=force_new,
    )


@app.command(help="Stop Gaiaflow production-like services.")
def stop(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.BaseAction.STOP,
    )


@app.command(help="Restart Gaiaflow production-like services.")
def restart(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    force_new: bool = typer.Option(
        False,
        "--force-new",
        "-f",
        help="Set this to true if you need a "
        "fresh production-like environment installation. ",
    ),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.BaseAction.RESTART,
    )


@app.command(help="Containerize your package into a docker image inside the "
                  "minikube cluster.")
def dockerize(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    image_name: str = typer.Option(DEFAULT_IMAGE_NAME, "--image-name", "-i",
                                   help=("Name of your image.")),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.ExtendedAction.DOCKERIZE,
        local=False,
        image_name=image_name
    )


@app.command(
    help="Create a config file for Airflow to talk to Kubernetes "
    "cluster. To be used only when debugging required."
)
def create_config(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.ExtendedAction.CREATE_CONFIG,
    )


@app.command(help="Create secrets to provide to the production-like environment.")
def create_secret(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
    name: str = typer.Option(..., "--name", help="Name of the secret"),
    data: list[str] = typer.Option(
        ..., "--data", help="Secret data as key=value pairs"
    ),
):
    imports = load_imports()
    secret_data = imports.parse_key_value_pairs(data)
    print(secret_data, name)
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action=imports.ExtendedAction.CREATE_SECRET,
        secret_name=name,
        secret_data=secret_data,
    )


@app.command(
    help="Clean Gaiaflow production-like services. This will only remove the "
    "minikube speicifc things. To remove local docker stuff, use the dev mode."
)
def cleanup(
    project_path: Path = typer.Option(..., "--path", "-p", help="Path to your project"),
):
    imports = load_imports()
    gaiaflow_path, user_project_path = imports.create_gaiaflow_context_path(
        project_path
    )
    gaiaflow_path_exists = imports.gaiaflow_path_exists_in_state(gaiaflow_path, True)
    if not gaiaflow_path_exists:
        typer.echo("Please create a project with Gaiaflow before running this command.")
        return
    imports.MinikubeManager.run(
        gaiaflow_path=gaiaflow_path,
        user_project_path=user_project_path,
        action="cleanup",
    )



import json
import os
import platform
import shutil
import socket
import subprocess
from pathlib import Path

import fsspec
import psutil
from ruamel.yaml import YAML

from gaiaflow.constants import (AIRFLOW_SERVICES, GAIAFLOW_STATE_FILE,
                                MINIO_SERVICES, MLFLOW_SERVICES, Action,
                                BaseAction, Service)
from gaiaflow.managers.base_manager import BaseGaiaflowManager
from gaiaflow.managers.utils import (create_directory, delete_project_state,
                                     find_python_packages,
                                     gaiaflow_path_exists_in_state,
                                     handle_error, log_error, log_info, run,
                                     save_project_state, set_permissions, convert_crlf_to_lf)

_IMAGES = [
    "docker-compose-airflow-apiserver:latest",
    "docker-compose-airflow-scheduler:latest",
    "docker-compose-airflow-dag-processor:latest",
    "docker-compose-airflow-triggerer:latest",
    "docker-compose-airflow-init:latest",
    "docker-compose-mlflow:latest",
    "minio/mc:latest",
    "minio/minio:latest",
    "postgres:13",
]

_VOLUMES = [
    "docker-compose_postgres-db-volume-airflow",
    "docker-compose_postgres-db-volume-mlflow",
]


class MlopsManager(BaseGaiaflowManager):
    """Manager class to Start/Stop/Restart MLOps Docker services."""

    def __init__(
        self,
        gaiaflow_path: Path,
        user_project_path: Path,
        action: Action,
        service: Service = Service.all,
        cache: bool = False,
        jupyter_port: int = 8895,
        delete_volume: bool = False,
        docker_build: bool = False,
        force_new: bool = False,
        prune: bool = False,
        prod_local: bool = False,
        **kwargs,
    ):
        if kwargs:
            raise TypeError(f"Unexpected keyword arguments: {list(kwargs.keys())}")
        self.service = service
        self.cache = cache
        self.jupyter_port = jupyter_port
        self.delete_volume = delete_volume
        self.docker_build = docker_build
        self.os_type = platform.system().lower()
        self.project_root = Path(__file__).resolve().parent
        self.fs = fsspec.filesystem("file")
        self.prod_local = prod_local

        super().__init__(
            gaiaflow_path=gaiaflow_path,
            user_project_path=user_project_path,
            action=action,
            force_new=force_new,
            prune=prune,
        )

    @classmethod
    def run(cls, **kwargs):
        action = kwargs.get("action")
        if action is None:
            raise ValueError("Missing required argument 'action'")

        manager = MlopsManager(**kwargs)

        action_map = {
            BaseAction.START: manager.start,
            BaseAction.STOP: manager.stop,
            BaseAction.RESTART: manager.restart,
            BaseAction.CLEANUP: manager.cleanup,
        }

        try:
            action_method = action_map[action]
        except KeyError:
            raise ValueError(f"Unknown action: {action}")

        action_method()

    def start(self):
        log_info("Starting Gaiaflow services")

        if self.force_new:
            self.cleanup()

        gaiaflow_path_exists = gaiaflow_path_exists_in_state(self.gaiaflow_path, True)

        if not gaiaflow_path_exists:
            log_info("Setting up directories...")
            create_directory(f"{self.user_project_path}/logs")
            create_directory(f"{self.user_project_path}/data")

            log_info("Updating .env file...")
            self._update_env_file_with_airflow_uid(f"{self.user_project_path}/.env")

            log_info("Creating gaiaflow context...")
            self._create_gaiaflow_context()

            log_info("Updating gaiaflow context with user project information...")
            self._update_files()

            save_project_state(self.user_project_path, self.gaiaflow_path)
        else:
            log_info(
                "Gaiaflow project already exists at "
                f"{self.gaiaflow_path}, "
                "skipping creating new context."
            )

        if self.service == Service.jupyter or self.service == Service.all:
            self._check_port()

        if self.docker_build:
            build_cmd = ["build"]
            if not self.cache:
                build_cmd.append("--no-cache")

            log_info("Building Docker images")
            self._docker_compose_action(build_cmd, self.service)

        if self.service == Service.all:
            self._start_jupyter()
            self._docker_compose_action(["up", "-d"], service=None)
        elif self.service == Service.jupyter:
            self._start_jupyter()
        else:
            self._docker_compose_action(["up", "-d"], service=self.service)

    def stop(self):
        log_info("Shutting down Gaiaflow services...")
        if self.service == Service.jupyter:
            self._stop_jupyter()
        elif self.service == Service.all:
            down_cmd = ["down"]
            if self.delete_volume:
                log_info("Removing volumes with shutdown")
                down_cmd.append("-v")
            self._docker_compose_action(down_cmd)
            self._stop_jupyter()
        else:
            down_cmd = ["down"]
            if self.delete_volume:
                log_info("Removing volumes with shutdown")
                down_cmd.append("-v")
            self._docker_compose_action(down_cmd, self.service)

        log_info("Stopped Gaiaflow services successfully")

    def _check_port(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            if sock.connect_ex(("127.0.0.1", self.jupyter_port)) == 0:
                handle_error(f"Port {self.jupyter_port} is already in use.")

    def _stop_jupyter(self):
        log_info(f"Attempting to stop Jupyter processes on port {self.jupyter_port}")
        for proc in psutil.process_iter(attrs=["pid", "name", "cmdline"]):
            try:
                cmdline = proc.info.get("cmdline") or []
                name = proc.info.get("name") or ""
                if any("jupyter-lab" in arg for arg in cmdline) or "jupyter" in name:
                    log_info(f"Terminating process {proc.pid} ({name})")
                    proc.terminate()
                    proc.wait(timeout=5)
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                continue

    @staticmethod
    def _docker_services_for(component):
        services = {
            "airflow": AIRFLOW_SERVICES,
            "mlflow": MLFLOW_SERVICES,
            "minio": MINIO_SERVICES,
        }
        return services.get(component, [])

    def _docker_compose_action(self, actions, service=None):
        if self.prod_local:
            base_cmd = [
                "docker",
                "compose",
                "-f",
                f"{self.gaiaflow_path}/docker_stuff/docker-compose/docker-compose.yml",
                "-f",
                f"{self.gaiaflow_path}/docker_stuff/docker-compose/docker-compose-minikube-network.yml",
            ]
        else:
            base_cmd = [
                "docker",
                "compose",
                "-f",
                f"{self.gaiaflow_path}/docker_stuff/docker-compose/docker-compose.yml",
            ]
        if service:
            services = MlopsManager._docker_services_for(service)
            print("Services:::", services, service)
            if not services:
                handle_error(f"Unknown service: {service}")
            cmd = base_cmd + actions + services
        else:
            cmd = base_cmd + actions

        log_info(f"Running: {' '.join(cmd)}")
        run(cmd, f"Error running docker compose {actions}")

    def _start_jupyter(self):
        log_info("Starting Jupyter Lab...")
        cmd = ["jupyter", "lab", "--ip=0.0.0.0", f"--port={self.jupyter_port}"]
        subprocess.Popen(cmd)

    def _update_env_file_with_airflow_uid(self, env_path):
        if self.os_type == "linux":
            uid = str(os.getuid())
        else:
            uid = 50000

        lines = []
        if os.path.exists(env_path):
            with open(env_path, "r") as f:
                lines = f.readlines()

        key_found = False
        new_lines = []
        for line in lines:
            if line.strip().startswith("AIRFLOW_UID="):
                new_lines.append(f"AIRFLOW_UID={uid}\n")
                key_found = True
            else:
                new_lines.append(line)

        if not key_found:
            new_lines.append(f"AIRFLOW_UID={uid}\n")

        with open(env_path, "w") as f:
            f.writelines(new_lines)

        log_info(f"Set AIRFLOW_UID={uid} in {env_path}")

    def _create_gaiaflow_context(self):
        self.fs.makedirs(self.gaiaflow_path, exist_ok=True)

        package_dir = Path(__file__).parent.parent.resolve()
        docker_dir = package_dir.parent / "docker_stuff"
        # docker_dir = package_dir / "docker_stuff"

        shutil.copytree(docker_dir, self.gaiaflow_path / "docker_stuff", dirs_exist_ok=True)
        shutil.copy(
            self.user_project_path / "environment.yml",
            self.gaiaflow_path / "environment.yml",
        )
        log_info(f"Gaiaflow context created at {self.gaiaflow_path}")

    def _update_files(self):
        yaml = YAML()
        yaml.preserve_quotes = True

        compose_path = (
            self.gaiaflow_path / "docker_stuff" / "docker-compose" / "docker-compose.yml"
        )

        with open(compose_path) as f:
            compose_data = yaml.load(f)

        x_common = compose_data.get("x-airflow-common", {})
        original_vols = x_common.get("volumes", [])
        new_volumes = []
        for vol in original_vols:
            parts = vol.split(":", 1)
            if len(parts) == 2:
                src, dst = parts
                src_path = (self.user_project_path / Path(src).name).resolve().as_posix()
                print("src_path", src_path)
                new_volumes.append(f"{src_path}:{dst}")

        existing_mounts = {Path(v.split(":", 1)[0]).name for v in new_volumes}
        python_packages = find_python_packages(self.user_project_path)

        for python_package in python_packages:
            set_permissions(python_package, 0o755)

        for child in self.user_project_path.iterdir():
            if (
                child.is_dir() and child.name not in existing_mounts and child.name
            ) in python_packages:
                dst_path = f"/opt/airflow/{child.name}"
                new_volumes.append(f"{child.resolve().as_posix()}:{dst_path}")

        kube_config_path = (self.gaiaflow_path.resolve() / "docker_stuff" / "kube_config_inline").as_posix()
        new_volumes.append(f"{kube_config_path}:/home/airflow/.kube/config")

        entrypoint_path = (
                    self.gaiaflow_path.resolve() / "docker_stuff" / "docker-compose" / "entrypoint.sh").as_posix()
        new_volumes.append(f"{entrypoint_path}:/opt/airflow/entrypoint.sh")

        # new_volumes.append(
        #     f"{self.gaiaflow_path.as_posix() / 'docker_stuff' / 'docker-compose'}/entrypoint.sh:/opt/airflow/entrypoint.sh"
        # )

        pyproject_path = (self.user_project_path.resolve()  / "pyproject.toml").as_posix()
        new_volumes.append(f"{pyproject_path}:/opt/airflow/pyproject.toml")

        new_volumes.append("/var/run/docker.sock:/var/run/docker.sock")

        compose_data["x-airflow-common"]["volumes"] = new_volumes

        with compose_path.open("w") as f:
            yaml.dump(compose_data, f)

        entrypoint_path = (
            self.gaiaflow_path / "docker_stuff" / "docker-compose" / "entrypoint.sh"
        )
        set_permissions(entrypoint_path)
        convert_crlf_to_lf(entrypoint_path)

    def cleanup(self):
        try:
            log_info("Attempting deleting Gaiaflow context at {self.gaiaflow_path}")
            shutil.rmtree(self.gaiaflow_path)
        except FileNotFoundError:
            log_error(f"Gaiaflow context not found at {self.gaiaflow_path}")
        try:
            log_info(
                f"Attempting deleting Gaiaflow project state at {GAIAFLOW_STATE_FILE}"
            )
            delete_project_state(self.gaiaflow_path)
        except (json.JSONDecodeError, FileNotFoundError):
            raise
        if self.prune:
            run(
                ["docker", "builder", "prune", "-a", "-f"],
                "Error pruning docker build cache",
            )
            run(
                ["docker", "system", "prune", "-a", "-f"],
                "Error pruning docker system",
            )
            run(
                ["docker", "volume", "prune", "-a", "-f"],
                "Error pruning docker volumes",
            )
            run(
                ["docker", "network", "rm", "docker-compose_ml-network"],
                "Error removing docker network",
            )
            for image in _IMAGES:
                run(["docker", "rmi", "-f", image], f"Error deleting image {image}")
            for volume in _VOLUMES:
                run(
                    ["docker", "volume", "rm", volume],
                    f"Error removing volume {volume}",
                )
        log_info("Gaiaflow cleanup complete!")



import os
import platform
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Set

import yaml

from gaiaflow.constants import (AIRFLOW_SERVICES, MINIO_SERVICES,
                                MLFLOW_SERVICES, Action, BaseAction,
                                ExtendedAction)
from gaiaflow.managers.base_manager import BaseGaiaflowManager
from gaiaflow.managers.mlops_manager import MlopsManager
from gaiaflow.managers.utils import (find_python_packages, log_error, log_info,
                                     run, set_permissions, is_wsl)


@contextmanager
def temporary_copy(src: Path, dest: Path):
    print("copying...", src, dest)
    shutil.copyfile(src, dest)
    try:
        yield
    finally:
        if dest.exists():
            dest.unlink()


class MinikubeManager(BaseGaiaflowManager):
    def __init__(
        self,
        gaiaflow_path: Path,
        user_project_path: Path,
        action: Action,
        force_new: bool = False,
        prune: bool = False,
        local: bool = False,
        image_name: str = "",
        **kwargs,
    ):
        # if kwargs:
        #     raise TypeError(f"Unexpected keyword arguments: {list(kwargs.keys())}")
        self.minikube_profile = "airflow"
        # TODO: get the docker image name automatically
        #  For CI, get the package name, version and create repository. See
        #  in test-airflow-ci test_ecr_push.yml
        self.os_type = platform.system().lower()
        self.local = local
        self.image_name = image_name

        super().__init__(
            gaiaflow_path=gaiaflow_path,
            user_project_path=user_project_path,
            action=action,
            force_new=force_new,
            prune=prune,
        )

    def _get_valid_actions(self) -> Set[Action]:
        return {
            BaseAction.START,
            BaseAction.STOP,
            BaseAction.RESTART,
            BaseAction.CLEANUP,
            ExtendedAction.DOCKERIZE,
            ExtendedAction.CREATE_CONFIG,
            ExtendedAction.CREATE_SECRET,
        }

    @classmethod
    def run(cls, **kwargs):
        action = kwargs.get("action")
        if action is None:
            raise ValueError("Missing required argument 'action'")

        manager = MinikubeManager(**kwargs)

        action_map = {
            BaseAction.START: manager.start,
            BaseAction.STOP: manager.stop,
            BaseAction.RESTART: manager.restart,
            BaseAction.CLEANUP: manager.cleanup,
            ExtendedAction.DOCKERIZE: manager.build_docker_image,
            ExtendedAction.CREATE_CONFIG: manager.create_kube_config_inline,
            ExtendedAction.CREATE_SECRET: manager.create_secrets,
        }

        try:
            action_method = action_map[action]
        except KeyError:
            raise ValueError(f"Unknown action: {action}")

        if action == ExtendedAction.CREATE_SECRET:
            action_method(kwargs["secret_name"], kwargs["secret_data"])
        else:
            action_method()

    def start(self):
        if self.force_new:
            self.cleanup()
        MlopsManager.run(gaiaflow_path=self.gaiaflow_path, user_project_path=self.user_project_path, action=BaseAction.STOP)
        log_info(f"Checking Minikube cluster [{self.minikube_profile}] status...")
        result = subprocess.run(
            ["minikube", "status", "--profile", self.minikube_profile],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        if b"Running" in result.stdout:
            log_info(f"Minikube cluster [{self.minikube_profile}] is already running.")
        else:
            log_info(
                f"Minikube cluster [{self.minikube_profile}] is not running. Starting..."
            )
            try:
                cmd = [
                        "minikube",
                        "start",
                        "--profile",
                        self.minikube_profile,
                        "--driver=docker",
                        "--cpus=4",
                        "--memory=4g",
                    ]
                if is_wsl():
                    cmd.append("--extra-config=kubelet.cgroup-driver=cgroupfs")
                run(
                    cmd,
                    f"Error starting minikube profile [{self.minikube_profile}]",
                )
            except subprocess.CalledProcessError:
                log_info("Cleaning up and starting again...")
                self.cleanup()
                self.start()

        self.create_kube_config_inline()
        MlopsManager.run(
            gaiaflow_path=self.gaiaflow_path,
            user_project_path=self.user_project_path,
            action=BaseAction.START,
            prod_local=True,
            force_new=self.force_new,
        )

    def stop(self):
        log_info(f"Stopping minikube profile [{self.minikube_profile}]...")
        try:
            run(
                ["minikube", "stop", "--profile", self.minikube_profile],
                f"Error stopping minikube profile [{self.minikube_profile}]",
            )
            log_info(f"Stopped minikube profile [{self.minikube_profile}]")
        except Exception as e:
            log_info(str(e))

    def create_kube_config_inline(self):
        kube_config = Path.home() / ".kube" / "config"
        backup_config = kube_config.with_suffix(".backup")
        filename = f"{self.gaiaflow_path / 'docker_stuff'}/kube_config_inline"

        if kube_config.exists():
            with open(kube_config, "r") as f:
                config_data = yaml.safe_load(f)

            with open(backup_config, "w") as f:
                yaml.dump(config_data, f)

                for cluster in config_data.get("clusters", []):
                    if self.os_type == "windows" and kube_config.exists():
                        log_info("Detected Windows: patching kube config with host.docker.internal")
                        server = cluster.get("cluster", {}).get("server", "")
                        if "127.0.0.1" in server or "localhost" in server:
                            cluster["cluster"]["server"] = server.replace(
                                "127.0.0.1", "host.docker.internal"
                            ).replace("localhost", "host.docker.internal")


                    elif is_wsl():
                        log_info("Detected WSL: patching kube config with minikube ip")
                        # ip = subprocess.check_output(["minikube", "ip", "-p", "airflow"], text=True).strip()
                        # cluster["cluster"]["server"] = f"https://{ip}:8443"
                        cluster["cluster"]["server"] = "https://192.168.49.2:8443"
                        cluster["cluster"]["insecure-skip-tls-verify"] = True

            with open(kube_config, "w") as f:
                yaml.dump(config_data, f)

        log_info("Creating kube config inline file...")
        with open(filename, "w") as f:
            subprocess.call(
                [
                    "minikube",
                    "kubectl",
                    "--",
                    "config",
                    "view",
                    "--flatten",
                    "--minify",
                    "--raw",
                ],
                cwd=self.gaiaflow_path / "docker_stuff",
                stdout=f,
            )

        log_info(f"Created kube config inline file {filename}")

        if self.os_type == "windows":
            log_info(
                f"Adding insecure-skip-tls-verfiy for local setup in kube config inline file {filename}"
            )
            with open(filename, "r") as f:
                kube_config_data = yaml.safe_load(f)


            for cluster in kube_config_data.get("clusters", []):
                cluster_data = cluster.get("cluster", {})
                if "insecure-skip-tls-verify" not in cluster_data:
                    cluster_data["insecure-skip-tls-verify"] = True

            log_info(f"Saving kube config inline file {filename}")
            with open(filename, "w") as f:
                yaml.safe_dump(kube_config_data, f, default_flow_style=False)

        if (self.os_type == "windows" or is_wsl()) and backup_config.exists():
            shutil.copy(backup_config, kube_config)
            backup_config.unlink()
            log_info("Reverted kube config to original state.")

    @staticmethod
    def _add_copy_statements_to_dockerfile(
        dockerfile_path: str, local_packages: list[str]
    ):
        with open(dockerfile_path, "r") as f:
            lines = f.readlines()

        env_index = next(
            (i for i, line in enumerate(lines) if line.strip().startswith("ENV")),
            None,
        )

        if env_index is None:
            raise ValueError("No ENV found in Dockerfile.")

        entrypoint_index = next(
            (
                i
                for i, line in enumerate(lines)
                if line.strip().startswith("ENTRYPOINT")
            ),
            None,
        )

        if entrypoint_index is None:
            raise ValueError("No ENTRYPOINT found in Dockerfile.")

        copy_lines = [f"COPY {pkg} ./{pkg}\n" for pkg in local_packages]
        copy_lines.append("COPY runner.py ./runner.py\n")

        updated_lines = (
            lines[: env_index + 1]
            + copy_lines  #
            + lines[entrypoint_index:]
        )
        with open(dockerfile_path, "w") as f:
            f.writelines(updated_lines)

        print("Dockerfile updated with COPY statements.")

    def build_docker_image(self):
        dockerfile_path = self.gaiaflow_path / "docker_stuff" / "user-package" / "Dockerfile"
        if not (dockerfile_path.exists()):
            log_error(f"Dockerfile not found at {dockerfile_path}")
            return

        log_info(f"Updating dockerfile at {dockerfile_path}")
        MinikubeManager._add_copy_statements_to_dockerfile(
            dockerfile_path, find_python_packages(self.user_project_path)
        )
        runner_src = Path(__file__).parent.parent.resolve() / "core" / "runner.py"
        runner_dest = self.user_project_path / "runner.py"

        with temporary_copy(runner_src, runner_dest):
            if self.local:
                log_info(f"Building Docker image [{self.image_name}] locally")
                run(
                    [
                        "docker",
                        "build",
                        "-t",
                        self.image_name,
                        "-f",
                        dockerfile_path,
                        self.user_project_path,
                    ],
                    "Error building docker image.",
                )
                # TODO: For windows?
                set_permissions("/var/run/docker.sock", 0o666)
            else:
                log_info(
                    f"Building Docker image [{self.image_name}] in minikube context"
                )
                result = subprocess.run(
                    [
                        "minikube",
                        "-p",
                        self.minikube_profile,
                        "docker-env",
                        "--shell",
                        "bash",
                    ],
                    stdout=subprocess.PIPE,
                    check=True,
                )
                env = os.environ.copy()
                for line in result.stdout.decode().splitlines():
                    if line.startswith("export "):
                        try:
                            key, value = line.replace("export ", "").split("=", 1)
                            env[key.strip()] = value.strip('"')
                        except ValueError:
                            continue
                run(
                    [
                        "docker",
                        "build",
                        "-t",
                        self.image_name,
                        "-f",
                        dockerfile_path,
                        self.user_project_path,
                    ],
                    "Error building docker image inside minikube cluster.",
                    env=env,
                )

    def create_secrets(self, secret_name: str, secret_data: dict[str, Any]):
        log_info(f"Checking if secret [{secret_name}] exists...")
        check_cmd = [
            "minikube",
            "kubectl",
            "-p",
            self.minikube_profile,
            "--",
            "get",
            "secret",
            secret_name,
        ]
        result = subprocess.run(
            check_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        if result.returncode == 0:
            log_info(f"Secret [{secret_name}] already exists. Skipping creation.")
        else:
            log_info(f"Creating secret [{secret_name}]...")
            create_cmd = [
                "minikube",
                "kubectl",
                "-p",
                self.minikube_profile,
                "--",
                "create",
                "secret",
                "generic",
                secret_name,
            ]
            for k, v in secret_data.items():
                create_cmd.append(f"--from-literal={k}={v}")
            subprocess.check_call(create_cmd)

    def cleanup(self):
        log_info(f"Deleting minikube profile: {self.minikube_profile}")
        run(
            ["minikube", "delete", "--profile", self.minikube_profile],
            f"Error deleting minikube profile [{self.minikube_profile}]",
        )
        for service in AIRFLOW_SERVICES + MLFLOW_SERVICES + MINIO_SERVICES:
            run(
                ["docker", "network", "disconnect", self.minikube_profile, service],
                f"Error disconnecting network from service: {service}",
            )
        run(
            ["docker", "network", "rm", "-f", "airflow"],
            "Error removing  airflow docker network",
        )
        log_info("Minikube Cleanup complete")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Set

from gaiaflow.constants import Action, BaseAction


class BaseGaiaflowManager(ABC):
    def __init__(
        self,
        gaiaflow_path: Path,
        user_project_path: Path,
        action: Action,
        force_new: bool = False,
        prune: bool = False,
    ):
        valid_actions = self._get_valid_actions()
        if action not in valid_actions:
            valid_names = sorted([a.name for a in valid_actions])
            raise ValueError(f"Invalid action '{action.name}'. Valid: {valid_names}")

        self.gaiaflow_path = gaiaflow_path
        self.user_project_path = user_project_path
        self.action = action
        self.force_new = force_new
        self.prune = prune

    @abstractmethod
    def start(self):
        """Start the services provided by the manager.

        It can use the `force_new` variable to start a fresh set of services
        after removing the old ones.
        """

    @abstractmethod
    def stop(self):
        """Stop the services provided by the manager."""

    def restart(self):
        """Restart the services provided by the manager."""
        self.stop()
        self.start()

    @abstractmethod
    def cleanup(self):
        """Cleanup the services provided by the manager.

        It can use the `prune` flag to permanently delete the services
        provided by the manager.
        """

    def _get_valid_actions(self) -> Set[Action]:
        """Return set of valid actions for this manager. Override in subclasses."""
        return {
            BaseAction.START,
            BaseAction.STOP,
            BaseAction.RESTART,
            BaseAction.CLEANUP,
        }



import json
import platform
from datetime import datetime

from airflow.providers.cncf.kubernetes.operators.pod import \
    KubernetesPodOperator
from airflow.providers.docker.operators.docker import DockerOperator
from airflow.providers.standard.operators.python import ExternalPythonOperator
from kubernetes.client import V1ResourceRequirements

from gaiaflow.constants import (DEFAULT_MINIO_AWS_ACCESS_KEY_ID,
                                DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY,
                                RESOURCE_PROFILES)

from .utils import build_env_from_secrets, inject_params_as_env_vars


class FromTask:
    def __init__(self, task: str, key: str = "return_value"):
        self.task = task
        self.key = key

    def to_dict(self) -> dict:
        return {"task": self.task, "key": self.key}


def split_args_kwargs(func_args=None, func_kwargs=None):
    func_args = func_args or []
    func_kwargs = func_kwargs or {}

    user_args = []
    xcom_args = {}
    for idx, arg in enumerate(func_args):
        if isinstance(arg, FromTask):
            xcom_args[str(idx)] = arg.to_dict()
        else:
            user_args.append(arg)

    user_kwargs = {}
    xcom_kwargs = {}
    for k, v in func_kwargs.items():
        if isinstance(v, FromTask):
            xcom_kwargs[k] = v.to_dict()
        else:
            user_kwargs[k] = v

    return user_args, xcom_args, user_kwargs, xcom_kwargs


class BaseTaskOperator:
    def __init__(
        self,
        task_id: str,
        func_path: str,
        func_args: list,
        func_kwargs: dict,
        image: str | None,
        secrets: list[str] | None,
        env_vars: dict,
        retries: int,
        params: dict,
        mode: str,
    ):
        self.task_id = task_id
        self.func_path = func_path
        self.image = image
        self.secrets = secrets
        self.env_vars = env_vars
        self.retries = retries
        self.params = params
        self.mode = mode

        (
            self.func_args,
            self.func_args_from_tasks,
            self.func_kwargs,
            self.func_kwargs_from_tasks,
        ) = split_args_kwargs(func_args, func_kwargs)

        self.func_args_from_tasks = self.func_args_from_tasks or {}
        self.func_kwargs_from_tasks = self.func_kwargs_from_tasks or {}

        self.func_kwargs_from_tasks = {
            k: (v if isinstance(v, dict) and "task" in v else FromTask(v).to_dict())
            for k, v in self.func_kwargs_from_tasks.items()
        }

    def create_task(self):
        raise NotImplementedError

    def _resolve_xcom_value(self, from_task_config):
        task_id = from_task_config["task"]
        key = from_task_config.get("key", "return_value")

        if key == "return_value":
            return f"{{{{ ti.xcom_pull(task_ids='{task_id}') }}}}"
        return f"{{{{ ti.xcom_pull(task_ids='{task_id}')['{key}'] }}}}"

    def resolve_args_kwargs(self):
        resolved_args = self.func_args or []
        resolved_kwargs = self.func_kwargs or {}

        for index_str, from_task_config in (self.func_args_from_tasks or {}).items():
            index = int(index_str)
            while len(resolved_args) <= index:
                resolved_args.append(None)
            resolved_args[index] = self._resolve_xcom_value(from_task_config)

        for k, from_task_config in (self.func_kwargs_from_tasks or {}).items():
            resolved_kwargs[k] = self._resolve_xcom_value(from_task_config)

        return resolved_args, resolved_kwargs

    def create_func_env_vars(self):
        final_args, final_kwargs = self.resolve_args_kwargs()
        return {
            "FUNC_PATH": self.func_path,
            "FUNC_ARGS": json.dumps(final_args),
            "FUNC_KWARGS": json.dumps(final_kwargs),
        }


class DevTaskOperator(BaseTaskOperator):
    def create_task(self):
        from gaiaflow.core.runner import run

        args, kwargs = self.resolve_args_kwargs()
        kwargs["params"] = self.params
        op_kwargs = {"func_path": self.func_path, "args": args, "kwargs": kwargs}

        return ExternalPythonOperator(
            task_id=self.task_id,
            python="/home/airflow/.local/share/mamba/envs/default_user_env/bin/python",
            python_callable=run,
            op_kwargs=op_kwargs,
            do_xcom_push=True,
            retries=self.retries,
        )


class ProdLocalTaskOperator(BaseTaskOperator):
    def create_task(self):
        if not self.image:
            raise ValueError("Docker image must be provided for Kubernetes tasks.")

        os_type = platform.system().lower()

        minikube_gateway = "NOTSET"

        if os_type == "linux":
            minikube_gateway = "192.168.49.1"
        elif os_type == "windows":
            minikube_gateway = "host.docker.internal"

        # If the user provides a gateway, that takes precedence.
        if "MINIKUBE_GATEWAY" in self.env_vars:
            minikube_gateway = self.env_vars.get("MINIKUBE_GATEWAY")

        mlflow_env_vars = {
            "MLFLOW_TRACKING_URI": f"http://{minikube_gateway}:5000",
            "MLFLOW_S3_ENDPOINT_URL": f"http://{minikube_gateway}:9000",
        }

        aws_access_key_id = DEFAULT_MINIO_AWS_ACCESS_KEY_ID
        if "AWS_ACCESS_KEY_ID" in self.env_vars:
            aws_access_key_id = self.env_vars.pop("AWS_ACCESS_KEY_ID")

        aws_secret_access_key = DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY
        if "AWS_SECRET_ACCESS_KEY" in self.env_vars:
            aws_secret_access_key = self.env_vars.pop("AWS_SECRET_ACCESS_KEY")

        minio_env_vars = {
            "AWS_ACCESS_KEY_ID": aws_access_key_id,
            "AWS_SECRET_ACCESS_KEY": aws_secret_access_key,
        }

        all_env_vars = {
            "MODE": self.mode.value,
            **self.env_vars,
            **inject_params_as_env_vars(self.params),
            **mlflow_env_vars,
            **minio_env_vars,
            **self.create_func_env_vars(),
        }
        env_from = build_env_from_secrets(self.secrets or [])

        profile_name = self.params.get("resource_profile", "low")
        profile = RESOURCE_PROFILES.get(profile_name)
        if profile is None:
            raise ValueError(f"Unknown resource profile: {profile_name}")

        resources = V1ResourceRequirements(
            requests={
                "cpu": profile["request_cpu"],
                "memory": profile["request_memory"],
            },
            limits={
                "cpu": profile["limit_cpu"],
                "memory": profile["limit_memory"],
                # "gpu": profile.get["limit_gpu"],
            },
        )

        return KubernetesPodOperator(
            task_id=self.task_id,
            image=self.image,
            cmds=["python", "-m", "runner"],
            env_vars=all_env_vars,
            env_from=env_from,
            get_logs=True,
            is_delete_operator_pod=True,
            log_events_on_failure=True,
            in_cluster=(self.mode == self.mode.PROD),
            do_xcom_push=True,
            retries=self.retries,
            params=self.params,
            container_resources=resources,
        )


class ProdTaskOperator(ProdLocalTaskOperator):
    """"""


class DockerTaskOperator(ProdLocalTaskOperator):
    def create_task(self):
        mlflow_tracking_uri = "http://mlflow:5000"
        if "MLFLOW_TRACKING_URI" in self.env_vars:
            mlflow_tracking_uri = self.env_vars.pop("MLFLOW_TRACKING_URI")

        mlflow_s3_endpoint_url = "http://minio:9000"
        if "MLFLOW_S3_ENDPOINT_URL" in self.env_vars:
            mlflow_s3_endpoint_url = self.env_vars.pop("MLFLOW_S3_ENDPOINT_URL")

        mlflow_env_vars = {
            "MLFLOW_TRACKING_URI": mlflow_tracking_uri,
            "MLFLOW_S3_ENDPOINT_URL": mlflow_s3_endpoint_url,
        }

        aws_access_key_id = DEFAULT_MINIO_AWS_ACCESS_KEY_ID
        if "AWS_ACCESS_KEY_ID" in self.env_vars:
            aws_access_key_id = self.env_vars.pop("AWS_ACCESS_KEY_ID")

        aws_secret_access_key = DEFAULT_MINIO_AWS_SECRET_ACCESS_KEY
        if "AWS_SECRET_ACCESS_KEY" in self.env_vars:
            aws_secret_access_key = self.env_vars.pop("AWS_SECRET_ACCESS_KEY")

        minio_env_vars = {
            "AWS_ACCESS_KEY_ID": aws_access_key_id,
            "AWS_SECRET_ACCESS_KEY": aws_secret_access_key,
        }

        combined_env = {
            "MODE": self.mode.value,
            **self.create_func_env_vars(),
            **self.env_vars,
            **inject_params_as_env_vars(self.params),
            **mlflow_env_vars,
            **minio_env_vars,
        }

        safe_image_name = self.image.replace(":", "_").replace("/", "_")

        return DockerOperator(
            task_id=self.task_id,
            image=self.image,
            container_name=safe_image_name
            + "_"
            + self.task_id
            + "_"
            + datetime.now().strftime("%Y%m%d%H%M%S")
            + "_container",
            api_version="auto",
            auto_remove="success",
            command=["python", "-m", "runner"],
            docker_url="unix://var/run/docker.sock",
            # docker_url="tcp://host.docker.internal:2375",
            environment=combined_env,
            network_mode="docker-compose_ml-network",
            mount_tmp_dir=False,
            do_xcom_push=True,
            retrieve_output=True,
            retrieve_output_path="/tmp/script.out",
            xcom_all=False,
        )



# !!PLEASE DO NOT EDIT/DELETE THIS!!
# This file is the main entry point to your package when using Airflow to
# run your tasks from your DAGs called from the task_factory.
# It imports the required function from your package and executes it with the
# arguments provided

import json
import os
import pickle
from typing import Any


def _extract_params_from_env(prefix="PARAMS_") -> dict[str, str]:
    return {
        k[len(prefix) :].lower(): v
        for k, v in os.environ.items()
        if k.startswith(prefix)
    }


def run(
    func_path: str | None = None,
    args: list | None = None,
    kwargs: dict[str, Any] | None = None,
) -> dict[str, str]:
    mode = os.environ.get("MODE", "dev")
    print(f"## Runner running in {mode} mode ##")
    if mode == "dev":
        print("args", args)
        print("kwargs", kwargs)
    else:
        func_path = os.environ.get("FUNC_PATH", "")
        args = json.loads(os.environ.get("FUNC_ARGS", "[]"))
        kwargs = json.loads(os.environ.get("FUNC_KWARGS", "{}"))
        params: dict = _extract_params_from_env()
        kwargs["params"] = params
        print("args", args)
        print("kwargs", kwargs)

    if not func_path:
        raise ValueError("func_path must be provided")

    module_path, func_name = func_path.rsplit(":", 1)
    import importlib

    module = importlib.import_module(module_path)
    func = getattr(module, func_name)

    print(f"Running {func_path} with args: {args} and kwargs :{kwargs}")
    result = func(*args, **kwargs)
    print("Function result:", result)
    print("mode::::", mode, type(mode))
    if mode == "prod" or mode == "prod_local":
        # This is needed when we use KubernetesPodOperator and want to
        # share information via XCOM.
        _write_xcom_result(result)
    if mode == "dev_docker":
        print("inside dev_docker condition")
        with open("/tmp/script.out", "wb+") as tmp:
            pickle.dump(result, tmp)
        # print("printing result now:::")
        # print(json.dumps(result))

    return result


def _write_xcom_result(result: Any) -> None:
    try:
        xcom_dir = "/airflow/xcom"
        os.makedirs(xcom_dir, exist_ok=True)

        with open(f"{xcom_dir}/return.json", "w") as f:
            json.dump(result, f)

        path = "/airflow/xcom/return.json"
        print("[DEBUG] File exists:", os.path.exists(path))
        print("[DEBUG] File size:", os.path.getsize(path))
        with open(path, "r") as f:
            print("[DEBUG] File contents:", f.read())

        print("Result written to XCom successfully")
    except Exception as e:
        print(f"Failed to write XCom result: {e}")
        raise


if __name__ == "__main__":
    run()
